#!/usr/bin/env python3
import dbus
import dbus.service
import shutil
import dbus.mainloop.glib
import subprocess
import re
import hashlib
import threading
from urllib.parse import urlparse
from gi.repository import GLib
import os
import sys

# ===== 配置区域 =====
MIN_FREE_BYTES = 9 * 1024 * 1024 * 1024  # 9GB
UPDATE_DIR = "/update"
DELTA_FILE = os.path.join(UPDATE_DIR, "update.xdelta")
DELTA_FILE_MD5 = os.path.join(UPDATE_DIR, "update.xdelta.md5")
DELTA_FILE_BUILDID = os.path.join(UPDATE_DIR, "update.xdelta.buildid")
SYSTEM_BUILDID = "/etc/kylin-build"
APPFS_PATH = "/appfs"
ROOTFS_DEVICE = "/dev/disk/by-partlabel/rootfs"
TEMP_SQFS = "/update/original.sqfs"
NEW_SQFS = "/update/new_system.sqfs"
UPDATE_SUCCESS_FILE = "/update/update_success"
DESKTOP_OPEN_SUCCESS = "/update/desktop_success"
ROOTFS_MOUNT_DIR = "/update/rootfs"
USR_DIFF_LIST = "/home/kylin/.log/match.log"
NEEDRECORVER = "/update/needrecorverrootfs"
OTAUPDATE_DEB = "ota-update-package"

# ===== D-Bus 服务定义 =====
class OtaUpdateService(dbus.service.Object):
    def __init__(self, bus, object_path='/com/kylin/OtaUpdate'):
        super().__init__(bus, object_path)
        self.cancel_flag = False

    @dbus.service.method("com.kylin.OtaUpdate",
                         in_signature='', out_signature='bi')
    def CanUpdateDetect(self):
        """
        判断是否可以执行升级操作。
        返回 (是否允许升级: bool, 状态码: int)
        
        状态码定义：
            0 - 本地差分包存在，可更新
            1 - 网络可用，但本地包不存在
            2 - 磁盘空间不足
        """
        if not self.DetectDiskSize():
            print("[CanUpdateDetect] 空间不足，升级失败")
            return False, 2

        if self.DetectDeltaFile():
            print("[CanUpdateDetect] 本地差分包存在")
            return True, 0

        result = self.DetectNetworkConnect()
        print(f"[CanUpdateDetect] 网络检测结果: {result}")
        return (True, 1) if result else (False, 1)

    def DetectDiskSize(self, min_bytes=MIN_FREE_BYTES):
        """
        检查 APPFS_PATH 的剩余空间是否 ≥ min_bytes。
        返回: bool
        """
        try:
            usage = shutil.disk_usage(APPFS_PATH)
            enough = usage.free >= min_bytes
            print(f"[DiskCheck] /appfs 剩余空间: {usage.free} 字节，足够: {enough}")
            return enough
        except FileNotFoundError:
            print(f"[DiskCheck][ERROR] 挂载点 {APPFS_PATH} 不存在")
            return False
        except Exception as e:
            print(f"[DiskCheck][ERROR] 检查磁盘空间失败: {e}")
            return False

    def DetectDeltaFile(self):
        """
        检查是否存在差分包 DELTA_FILE。
        返回: bool
        """
        exists = os.path.isfile(DELTA_FILE)
        print(f"[DeltaCheck] 差分包存在: {exists}")
        return exists

    def DetectNetworkConnect(self):
        """
        使用 ping 检查网络是否可达。
        返回: bool
        """
        try:
            #使用sources.list的源当做服务器地址
            ota_server = self.get_first_active_repo_host('/etc/apt/sources.list')
            result = subprocess.run(
                ['ping', '-c', '1', '-W', '1', ota_server],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
            is_ok = (result.returncode == 0)
            print(f"[NetCheck] ping {ota_server} 成功: {is_ok}")
            return is_ok
        except Exception as e:
            print(f"[NetCheck][ERROR] 网络检测异常: {e}")
            return False


    @dbus.service.method("com.kylin.OtaUpdate",
                         in_signature='', out_signature='bi')
    def CheckUpdateSystem(self):
        """
        升级系统。
        返回 (升级状态: bool, 状态码: int)
        
        状态码定义：
            true 0 - 可以升级,进入升级流程
            true 1 - 系统已是最新版本
            false  - 升级不可用
        """
        if not self.DetectDeltaFile():
            print("[UpdateSystem] 本地差分包不存在")
            return False, 0

        """防止更新完成后没有重启系统再次打开升级页面判断系统已更新"""
        if os.path.exists(UPDATE_SUCCESS_FILE):
            print("[UpdateSystem] 系统已更新")
            return True, 1

        result = self.CompareSystemVersion()
        print(f"[UpdateSystem] 系统版本对比结果: {result}")
        return (True, 0) if result else (True, 1)

    def CompareSystemVersion(self):
        """
        检查系统版本和差分包版本
        返回 (差分包版本高可以升级: True, 版本已是最新或者其他错误: False)
        """
        try:
            if not os.path.exists(DELTA_FILE_BUILDID):
                print("[CompareSystemVersion] 差分包 buildid 文件不存在")
                return False

            with open(SYSTEM_BUILDID, "r") as f:
                content = f.read()
            system_buildid = self.extract_buildid(content)

            with open(DELTA_FILE_BUILDID, "r") as f:
                delta_buildid_str = f.read().strip()

            if system_buildid is None:
                print("[CompareSystemVersion] 系统 buildid 无法解析")
                return False

            if not delta_buildid_str.isdigit():
                print(f"[CompareSystemVersion] 差分包 buildid 文件内容非法: {delta_buildid_str}")
                return False

            delta_buildid = int(delta_buildid_str)

            return delta_buildid > system_buildid

        except Exception as e:
            print(f"[CompareSystemVersion] 异常: {e}")
            return False

    def extract_buildid(self,text):
        match = re.search(r'buildid:\s*(\d+)', text)
        if match:
            return int(match.group(1))
        return None

    @dbus.service.method("com.kylin.OtaUpdate",
                         in_signature='', out_signature='')
    def RealUpdateSystem(self):
        if hasattr(self, 'update_thread') and self.update_thread.is_alive():
            self.UpgradeFailed("升级正在进行中，忽略重复请求")
            print("[RealUpdateSystem] 升级正在进行中，忽略重复请求")
            return

        self.cancel_flag = False
        self.update_thread = threading.Thread(target=self._do_update_work)
        self.update_thread.start()

    def _do_update_work(self):
        try:
            # 1. 检查升级文件
            if not os.path.isfile(DELTA_FILE):
                print("[RealUpdateSystem] 升级文件不存在")
                self.UpgradeFailed("差分文件不存在")
                return
            
            if not os.path.isfile(DELTA_FILE_MD5):
                print("[RealUpdateSystem] 升级文件 MD5 文件不存在")
                self.UpgradeFailed("MD5 文件不存在")
                return
            
            # 2. 校验 MD5
            if not self.verify_md5(DELTA_FILE, DELTA_FILE_MD5):
                print("[RealUpdateSystem] MD5 校验失败")
                self.UpgradeFailed("MD5 校验失败")
                return
            
            self.UpgradeStatus("差分包MD5 校验通过", 10)
            print("[RealUpdateSystem] MD5 校验通过")

            if self.cancel_flag:
                print("[RealUpdateSystem] 用户请求中止")
                self.CancelSuccess()
                return
            
            # 3. 使用 dd 拷贝当前 rootfs 为 sqfs 镜像（简化假设，可能实际需挂载和转格式）
            subprocess.check_call(['dcfldd', f'if={ROOTFS_DEVICE}', f'of={TEMP_SQFS}', 'bs=4M'])
            if self.cancel_flag:
                print("[RealUpdateSystem] 用户请求中止")
                # 清理临时文件
                if os.path.isfile(f'{TEMP_SQFS}'):
                    os.remove(f'{TEMP_SQFS}')
                self.CancelSuccess()
                return
            self.UpgradeStatus("当前系统已备份", 50)
            print("[RealUpdateSystem] 当前 rootfs 已备份")
            

            # 4. 应用升级补丁（例如用 xdelta3）
            subprocess.check_call(['xdelta3', '-d', '-s', TEMP_SQFS, DELTA_FILE, NEW_SQFS])
            if self.cancel_flag:
                print("[RealUpdateSystem] 用户请求中止")
                if os.path.isfile(f'{TEMP_SQFS}'):
                    os.remove(f'{TEMP_SQFS}')
                if os.path.isfile(f'{NEW_SQFS}'):
                    os.remove(f'{NEW_SQFS}')
                self.CancelSuccess()
                return
            else:
                self.CannotCancelUpdate()
            self.UpgradeStatus("升级补丁应用成功", 70)
            print("[RealUpdateSystem] 升级补丁应用成功")
            

            # 5. 将新系统写入 rootfs 分区
            subprocess.check_call(['dcfldd', f'if={NEW_SQFS}', f'of={ROOTFS_DEVICE}', 'bs=4M'])
            self.UpgradeStatus("系统更新完成", 90)
            print("[RealUpdateSystem] 新系统已写入 rootfs")

            with open(UPDATE_SUCCESS_FILE, "w") as f:
                pass  # 创建后立即关闭，不写内容

            # 6. 对比新系统和/appfs/usr下不同的文件并删除
            if not os.path.exists(ROOTFS_MOUNT_DIR):
                os.makedirs(ROOTFS_MOUNT_DIR)
            subprocess.check_call(['mount', ROOTFS_DEVICE, ROOTFS_MOUNT_DIR])
            subprocess.check_call(['userdata-diff-usr', f'{ROOTFS_MOUNT_DIR}/usr', f'{APPFS_PATH}/usr'])
            subprocess.check_call(['umount', ROOTFS_MOUNT_DIR])
            self.delete_files_from_list(USR_DIFF_LIST);

            # 7. 清除/var/lib/dpkg/status缓存文件，确保dpkg列表正确
            if os.path.isfile(f'{APPFS_PATH}/var/lib/dpkg/status'):
                os.remove(f'{APPFS_PATH}/var/lib/dpkg/status')
            
            # 8. 清理中间文件
            os.remove(NEW_SQFS)
            
            # 9. 通知客户端升级成功
            self.UpgradeSuccess()

        except subprocess.CalledProcessError as e:
            print(f"[RealUpdateSystem] 子进程出错: {e}")
            self.UpgradeFailed(f"子进程出错: {e}")
        except Exception as e:
            print(f"[RealUpdateSystem] 未知错误: {e}")
            self.UpgradeFailed(str(e))

    def verify_md5(self, file_path, md5_file_path):
        try:
            with open(md5_file_path, 'r') as f:
                expected_md5 = f.read().strip()
    
            hash_md5 = hashlib.md5()
            with open(file_path, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
    
            return hash_md5.hexdigest() == expected_md5
        except Exception as e:
            print(f"[verify_md5] 异常: {e}")
            return False

    def delete_files_from_list(self, filelist_path):
        with open(filelist_path, 'r') as f:
            for line in f:
                filepath = line.strip()
                if not filepath:
                    continue  # 跳过空行
                try:
                    if os.path.isfile(filepath):
                        os.remove(filepath)
                    else:
                        print(f"[跳过] 不是文件或不存在: {filepath}")
                except Exception as e:
                    print(f"[错误] 删除 {filepath} 失败: {e}")

    @dbus.service.method("com.kylin.OtaUpdate",
                         in_signature='', out_signature='')
    def DesktopOpenSuccess(self):
        #升级成功进入桌面，删除补丁文件，创建成功标志供ramdisk识别
        if os.path.isfile(f'{DELTA_FILE}'):
            os.remove(f'{DELTA_FILE}')
        if os.path.isfile(f'{DELTA_FILE_MD5}'):
            os.remove(f'{DELTA_FILE_MD5}')
        if os.path.isfile(f'{DELTA_FILE_BUILDID}'):
            os.remove(f'{DELTA_FILE_BUILDID}')
        if not os.path.isfile(f'{DESKTOP_OPEN_SUCCESS}'):
            with open(DESKTOP_OPEN_SUCCESS, "w") as f:
                pass  # 创建后立即关闭，不写内容

    @dbus.service.method("com.kylin.OtaUpdate",
                         in_signature='', out_signature='')
    def RecorverRootfs(self):
        #还原系统
        if os.path.isfile(f'{TEMP_SQFS}'):
            subprocess.check_call(['dcfldd', f'if={TEMP_SQFS}', f'of={ROOTFS_DEVICE}', 'bs=4M'])
            os.remove(f'{TEMP_SQFS}')
        if os.path.isfile(f'{NEEDRECORVER}'):
            os.remove(f'{NEEDRECORVER}')

    def get_first_active_repo_host(self, filepath):
         with open(filepath, 'r') as f:
             for line in f:
                 line = line.strip()
                 # 忽略空行和注释行
                 if not line or line.startswith('#'):
                     continue

                 # 找到 URL（从第一个以 http 开头的词开始）
                 parts = line.split()
                 for part in parts:
                     if part.startswith('http://') or part.startswith('https://'):
                         # 解析 URL，提取主机名
                         host = urlparse(part).hostname
                         return host
         return None

    @dbus.service.method("com.kylin.OtaUpdate",
                         in_signature='', out_signature='')
    def CheckUpdate(self):
        if not self.check_apt_update():
            return

        if not self.check_package_upgrade_simulation(f'[OTAUPDATE_DEB]'):
            return

        self.upgrade_package(f'[OTAUPDATE_DEB]')

    def check_apt_update(self):
         try:
             # 执行 apt-get update，-qq 表示安静模式（仅错误输出）
             result = subprocess.run(
                 ['sudo', 'apt-get', 'update', '-qq'],
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
                 text=True
             )

             # 检查返回码是否为 0
             if result.returncode == 0:
                 print("[check_apt_update] apt update 成功")
                 self.CheckUpdateStatus('软件源更新成功')
                 return True
             else:
                 print("[check_apt_update] apt update 失败")
                 print(result.stderr)
                 self.CheckUpdateFailed(result.stderr)
                 return False

         except Exception as e:
             print(f"运行出错: {e}")
             self.CheckUpdateFailed(f"运行出错: {e}");
             return False

    def check_package_upgrade_simulation(self, pkg_name):
         try:
             result = subprocess.run(
                 ['apt-get', '-s', 'dist-upgrade'],
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
                 text=True
             )

             if result.returncode != 0:
                 print("[check_package_upgrade_simulation] 执行 apt-get 模拟升级出错：")
                 print(result.stderr)
                 self.CheckUpdateFailed('执行 apt-get 模拟升级出错')
                 return False

             for line in result.stdout.splitlines():
                 if line.strip().startswith("Inst") and pkg_name in line:
                     print(f"[check_package_upgrade_simulation] 软件包 {pkg_name} 将被升级")
                     self.CheckUpdateStatus('存在可用升级,正在下载...')
                     return True

             print(f"[check_package_upgrade_simulation] 软件包 {pkg_name} 没有可用升级")
             self.CheckUpdateFailed('没有可用的升级包,请联系麒麟软件更换软件源')
             return False

         except Exception as e:
             print(f"运行出错: {e}")
             self.CheckUpdateFailed(f"运行出错: {e}")
             return False

    def upgrade_package(self, pkg_name):
         try:
             print(f"[upgrade_package] 开始升级软件包：{pkg_name}")
             result = subprocess.run(
                 ['sudo', 'apt-get', 'install', '-y', pkg_name],
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
                 text=True
             )

             if result.returncode == 0:
                 print(f"[upgrade_package] 软件包 {pkg_name} 升级成功")
                 self.CheckUpdateSuccess()
             else:
                 print(f"[upgrade_package] 软件包 {pkg_name} 升级失败")
                 print(result.stderr)
                 self.CheckUpdateFailed(result.stderr)

         except Exception as e:
             print(f"运行出错: {e}")
             self.CheckUpdateFailed(f"运行出错: {e}")

    @dbus.service.method("com.kylin.OtaUpdate", in_signature='', out_signature='')
    def CancelUpdateSystem(self):
        print("[CancelUpdate] 取消更新标志")
        self.cancel_flag = True

    @dbus.service.signal("com.kylin.OtaUpdate")
    def UpgradeSuccess(self):
        """通知客户端升级完成"""
        print("[Signal] UpgradeSuccess 已发送")
    
    @dbus.service.signal("com.kylin.OtaUpdate", signature='s')
    def UpgradeFailed(self, reason = ''):
        """通知客户端升级失败，并附带原因"""
        print(f"[Signal] UpgradeFailed 已发送: {reason}")

    @dbus.service.signal("com.kylin.OtaUpdate", signature='si')
    def UpgradeStatus(self, status = '', percentage = 0):
        """通知客户端升级过程，并附带进度"""
        print(f"[Signal] UpgradeStatus 已发送: {status}")

    @dbus.service.signal("com.kylin.OtaUpdate", signature='s')
    def CheckUpdateFailed(self, reason = ''):
        """通知客户端检查升级失败，并附带原因"""
        print(f"[Signal] CheckUpdateFailed 已发送: {reason}")

    @dbus.service.signal("com.kylin.OtaUpdate", signature='s')
    def CheckUpdateStatus(self, status = ''):
        """通知客户端检查升级过程，并附带进度"""
        print(f"[Signal] CheckUpdateStatus 已发送: {status}")

    @dbus.service.signal("com.kylin.OtaUpdate")
    def CheckUpdateSuccess(self):
        """通知客户端检查升级成功"""
        print(f"[Signal] CheckUpdateSuccess 已发送")

    @dbus.service.signal("com.kylin.OtaUpdate")
    def CancelSuccess(self):
        """通知客户端取消升级成功"""
        print("[Signal] CancelSuccess 已发送")

    @dbus.service.signal("com.kylin.OtaUpdate")
    def CannotCancelUpdate(self):
        """通知客户端无法取消升级"""
        print("[Signal] CannotCancelUpdate 已发送")

# ===== 主程序入口 =====
def main():
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    bus = dbus.SystemBus()

    try:
        name = dbus.service.BusName("com.kylin.OtaUpdate", bus, allow_replacement=False, replace_existing=False)
    except dbus.exceptions.NameExistsException:
        print("[Main] 服务 com.kylin.OtaUpdate 已经在运行，退出。")
        sys.exit(1)

    service = OtaUpdateService(bus)
    print("[Main] OtaUpdate D-Bus 服务已启动...")
    loop = GLib.MainLoop()
    loop.run()

if __name__ == '__main__':
    main()
